@ mode4ep.asm
@ mic, 2005 | micol972@gmail.com

@ Uses a rotoscale BG with an extended palette in mode 4

.text
.global main
.global forever


.equ PALETTE_RAM,	0x5000000
.equ PALETTE_RAM_SIZE,	0x0000400
.equ BG_RAM,		0x6000000
.equ BG_SUB_RAM, 	0x6200000
.equ LCDC_VRAM_E,	0x6880000

.equ DISPCNT,		0x4000000
.equ BG2CNT,		0x400000c
.equ BG3CNT,		0x400000e
.equ BG2PA,		0x4000020
.equ BG2X,		0x4000028
.equ BG3PA,		0x4000030
.equ VRAMCNT,		0x4000240
.equ VRAMCNT_E,		0x4000244
.equ POWERCNT,		0x4000304
.equ DISPCNT_SUB,	0x4001000
.equ BG0CNT_SUB,	0x4001008
.equ BG0HOFS_SUB,	0x4001010
.equ BG0VOFS_SUB,	0x4001012

.equ BG_MODE_0,		0x10000
.equ BG_MODE_4,		0x10004
.equ BG_MODE_6,		0x10006
.equ BG_EXTPAL_ENABLE,	0x40000000
.equ BG_SIZE_256x256,	(1<<14)
.equ SCREENBASE_SHIFT,	8

.equ BG0_ENABLE,	0x0100
.equ BG1_ENABLE,	0x0200
.equ BG2_ENABLE,	0x0400
.equ BG3_ENABLE,	0x0800
.equ OBJ_ENABLE,	0x1000

.equ BANK_A_BG,		0x81
.equ BANK_E_LCDC_6880000,0x80
.equ BANK_E_BG_EXTPAL_0123,0x84


main:

@ Write some junk to DISPCNT (just for fun..)
ldr r0,=0x11340
mov r1,#DISPCNT
str r0,[r1]


@ Turn on both screens, main screen on top
ldr r0,=0x830F
ldr r1,=POWERCNT
str r0,[r1]


@ Map VRAM bank A to main BG, bank C to sub BG
mov r0,#BANK_A_BG
ldr r1,=VRAMCNT
strb r0,[r1],#2
add r0,r0,#3
strb r0,[r1]


@ Set up rotation/scaling parameters for BG3
ldr r0,=BG3PA
mov r2,#0
mov r1,#0x100
strh r1,[r0],#2
strh r2,[r0],#2
strh r2,[r0],#2
strh r1,[r0],#2
ldr r2,=0x0
str r2,[r0],#4
ldr r2,=0x0
str r2,[r0],#4


@ Set up palette
mov r0,#0
mov r1,#PALETTE_RAM
add r9,r1,#PALETTE_RAM_SIZE
mov r7,#256
mov r8,#384
set_palette:
	mov r2,r0,lsr#1		@ 0..31
	rsb r3,r2,#31		@ 31..0
	strh r2,[r9]
	orr r4,r3,r2,lsl#5
	strh r4,[r9,#128]
	mov r4,r2,lsl#10
	orr r4,r4,r3,lsl#5
	strh r4,[r9,r7]
	mov r4,r3,lsl#10
	strh r4,[r1]
	strh r4,[r1,#128]
	strh r4,[r1,r7]
	strh r4,[r1,r8]
	strh r4,[r9,r8]
	add r1,r1,#2
	add r9,r9,#2
	add r0,r0,#1
	cmp r0,#64
	bne set_palette


@ Map bank E to LCDC
ldr r0,=VRAMCNT_E
mov r1,#BANK_E_LCDC_6880000
strb r1,[r0]

@ Copy palette data to slot 3
ldr r2,=palette
ldr r3,=(LCDC_VRAM_E + 0x6000)
mov r4,#4096
copy_pal:
	ldrh r5,[r2],#2
	strh r5,[r3],#2
	subs r4,r4,#1
	bne copy_pal

@ Map bank E to BG Ext Pltt
mov r1,#BANK_E_BG_EXTPAL_0123
strb r1,[r0]


@ Store some variables for the plasma at offset 0x100000 in RAM
ldr r0,=0x2100000
mov r8,#0
mov r9,#0
mov r10,#0
mov r11,#0
stmia r0!,{r8-r11}

@ Store BG0 and BG1 scrolling variables
stmia r0,{r8-r11}


@ Create 1 tiles 
mov r0,#0
mov r1,#BG_RAM
ldr r4,=0x0000		@ Color 0
create_tiles:
	strh r2,[r1],#2
	add r0,r0,#1
	cmp r0,#32
	bne create_tiles

ldr r0,=tiles
mov r1,#BG_RAM
add r1,r1,#64
mov r2,#128
copy_tiles:
	ldrh r3,[r0],#2
	strh r3,[r1],#2
	subs r2,r2,#1
	bne copy_tiles
	
@ Now fill BG3 and Sub BG0 with some tiles	
ldr r0,=(BG_RAM + 0x4000)
mov r8,#0x0200000
mov r1,#0
ldr r5,=map
fill_bg_vert:
	mov r2,#0
	fill_bg_horz:
		and r3,r1,#1
		and r4,r2,#1
		orr r3,r4,r3,lsl#1
		add r3,r3,#1
		ldr r4,[r5,r2,lsl#2]
		orr r3,r3,r4,lsl#12
		strh r3,[r0]

		mov r3,#0
		strh r3,[r0,r8]		@ Clear Sub BG0
		add r0,r0,#2
		add r2,r2,#1
		cmp r2,#32
		bne fill_bg_horz
	add r1,r1,#1
	cmp r1,#32
	bne fill_bg_vert


@ Copy font data
ldr r0,=font
ldr r1,=BG_SUB_RAM
ldr r2,=2688
copy_font:
	ldrh r3,[r0],#2
	strh r3,[r1],#2
	subs r2,r2,#1
	bne copy_font
	

@ Set BG mode 4 and enable BG3 and extended palettes
ldr r0,=(BG_MODE_4 | BG_EXTPAL_ENABLE | BG3_ENABLE)
mov r1,#DISPCNT
str r0,[r1]	

@ BG3: ext pltt mode, 256x256 pixels, screen base 0x4000, priority 0
ldr r0,=(BG_SIZE_256x256 | (8<<SCREENBASE_SHIFT))
ldr r1,=BG3CNT
strh r0,[r1]



@@@@ Sub core setup @@@@

@ Set mode 0 and enable BG0
ldr r0,=(BG_MODE_0 | BG0_ENABLE)
ldr r1,=DISPCNT_SUB
str r0,[r1]	

@ Sub BG0: text bg, 256 colors, 256x256 pixels, screen base 0x4000, priority 0
ldr r0,=0x0880
ldr r1,=0x4001008
strh r0,[r1]

@ Print some text on Sub BG0
bl VSync
ldr r0,=szText
mov r1,#48
mov r2,#88
mov r3,#0x4000
bl DrawText



@@@@ Main loop @@@@


forever:
	bl VSync

	b forever



VSync:
  stmfd sp!,{r0-r1}
  mov r0,#DISPCNT
  add r0,r0,#4		@ DISPCNT+4 = DISPSTAT

  vs_loop1:
    ldrh r1,[r0]
    ands r1,r1,#1
  bne vs_loop1

  vs_loop2:
    ldrh r1,[r0]
    ands r1,r1,#1
  beq vs_loop2

  ldmfd sp!,{r0-r1}
  mov pc,lr


DrawText:
@ r0: lpszText
@ r1: x
@ r2: y
@ r3: screen base
	stmfd sp!,{r4-r10}

	mov r9,r2,lsr#3
	mov r9,r9,lsl#6
	mov r7,r1,lsr#3
	add r9,r9,r3
	add r9,r9,r7,lsl#1
	add r9,r9,#0x6000000
	add r9,r9,#0x0200000
	
dt_cloop:	
	ldrb r4,[r0],#1
	cmp r4,#0
	beq dt_null
	cmp r4,#32
	moveq r4,#0
	subne r4,r4,#37

	strh r4,[r9],#2	
	
	b dt_cloop
dt_null:
	ldmfd sp!,{r4-r10}
	mov pc,lr

.pool
.align



.align 2
.pool

.align 2
font:
	.incbin "font_grad.pat"

.align 2
tiles:
	.incbin "eptiles.pat"
.align 2
map:
	.word 0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,14,14,15,15
	
.align 2	
palette:
	.incbin "eptiles.pal"
 
 .align 2
 szText: .asciz "MODE 4 EXT PLTT DEMO" 

.align
.pool
.end
